function [truth_train, Y_train, X_train] = generate_synthetic_data(N_l, N_u)

% function [Y_train, X_train, truth_train] = generate_synthetic_data(N_l, N_u)
%
% Function creates synthetic training set that can be used to train the
% model for semi-supervised aggregation of experts described in Djuric et
% al., 'Semi-Supervised Learning for Integration of Aerosol Predictions
% from Multiple Satellite Instruments', IJCAI 2013. The generated data has
% K = 5 experts and R = 2 data-generating regimes. The accuracies of
% experts in the two regimes are reversed, meaning that the first expert is
% the most accurate in regime no. 1 while being the least accurate in
% regime no. 2, and so on. Training examples are also described by
% two-dimensional (i.e., D = 2) inputs X_train, which are used by the
% training algorithm to split the training examples into two clusters.
%
% Inputs:
%   N_l         - number of labeled data in the training set
%   N_u         - number of unlabeled data in the training set
%
% Outputs:
%   truth_train - vector of length (N_l + N_u), holding ground-truth
%                 labels; NaN denotes missing label
%   Y_train     - (N x K) matrix, each row represents predictions of K = 5 
%                 experts for one training example; NaN denotes missing
%                 expert prediction
%   X_train     - (N x D) matrix, each row represents explanatory variables
%                 for the corresponding training example with D = 2
%  
% Author: Nemanja Djuric (nemanja@temple.edu)
% Date  : October 25th, 2013

% set variances of the experts for two regimes
sigmas1 = [0.1, 0.2, 0.3, 0.4, 0.5];
sigmas2 = [0.5, 0.4, 0.3, 0.2, 0.1];
% set the parameters for data features
num_features = 2;
feature_variance = 0.5;
mean1 = 1;
mean2 = -1;
K = length(sigmas1);

truth_train = randn(N_l + N_u, 1);          % ground-truth values
Y_train = zeros(N_l + N_u, K);              % expert predictions
X_train = zeros(N_l + N_u, num_features);   % example features
for i = 1 : N_l + N_u
    % flip a coin, and create example either from one or the other regime
    if (rand < 0.5)
        Y_train(i, :) = truth_train(i) + sigmas1 .* randn(1, K);
        X_train(i, :) = mean1 + feature_variance * randn(1, num_features);
    else
        Y_train(i, :) = truth_train(i) + sigmas2 .* randn(1, K);
        X_train(i, :) = mean2 + feature_variance * randn(1, num_features);
    end;
end;

% generate missing experts with 50% chance
missing_ind_train = rand(size(Y_train)) >= 0.5;
for i = 1 : N_u + N_l
    miss_ind = missing_ind_train(i, :);
    % if all experts are missing then take at least one as non-missing
    if (sum(miss_ind) == 5)
        miss_ind(ceil(K * rand)) = false;
        missing_ind_train(i, :) = miss_ind;
    end;
end;
% NaN entries denote missing data
Y_train(missing_ind_train) = NaN;
truth_train(1 : N_u) = NaN;