function agg_preds = semisuper_combination_of_experts_test(Sigma, q, Lambda, Y_test, X_test)

% function aggreg_preds = semisuper_combination_of_experts_test(Sigma, q, Lambda, Y_test, X_test)
%
% Function tests a regression model for semi-supervised combination of 
% experts. Given experts' predictions and explanatory variables for each
% test example, it outputs the aggregated predictions.
%
% Inputs:
%   Sigma       - cell array of length R, each element is the
%                 sigma matrix for one regime; see eq. (2) in the paper
%   q           - (R x D) matrix, each row is a prototype vector for one
%                 regime; see eq. (28)
%   Lambda      - cell array of length R, each element is a feature scaling
%                 matrix for one regime; see eq. (28)
%   Y_test      - (N x K) matrix, each row represents predictions of K 
%                 experts for one test example; NaN denotes missing expert
%                 prediction
%   X_test      - (N x D) matrix, each row represents explanatory variables
%                 for the corresponding test example; if R is larger than 1
%                 value of D needs to be larger than 0, otherwise if R = 1 
%                 you can leave this input empty
%
% Outputs:
%   agg_preds   - vector of length N, with aggregated experts' predictions;
%                 if all experts are missing the corresponding output value
%                 is equal to NaN
%  
% Author: Nemanja Djuric (nemanja@temple.edu), according to paper Djuric et
%         al., 'Semi-Supervised Learning for Integration of Aerosol
%         Predictions from Multiple Satellite Instruments', IJCAI 2013.
% Date  : October 25th, 2013

[N_test, K] = size(Y_test);
R = length(Sigma);
agg_preds = zeros(N_test, 1);
pi_ir = zeros(N_test, R);

if (nargin < 4) || (R == 1)
    X_test = zeros(N_test, 1);
end;

% find precision matrices to speed up the computation later
Sigmas_inv = cell(1, R);
for r = 1 : R
    Sigmas_inv{r} = inv(Sigma{r});
end;

% aggregate expert predictions for each test example
for i = 1 : N_test
    % compute pi_ir
    for r = 1 : R
        pi_ir(i, r) = exp(-(X_test(i, :) - q(r, :)) * Lambda{r} * (X_test(i, :) - q(r, :))');
    end;
    pi_ir(i, :) = pi_ir(i, :) / sum(pi_ir(i, :));
    
    % find the aggregated prediction
    agg_preds(i) = 0;
    for r = 1 : R
        agg_preds(i) = agg_preds(i) + pi_ir(i, r) * find_posterior_mean(Y_test(i, :), Sigmas_inv{r});
    end;
end;

function aggreg_pred = find_posterior_mean(expert_preds, Sigma_inv)
% compute marginalization stuff to handle missing data
miss_ind = isnan(expert_preds);
Q_mat = Sigma_inv(miss_ind, miss_ind);
V_mat = Sigma_inv(~miss_ind, miss_ind);
U_mat = Sigma_inv(~miss_ind, ~miss_ind);
U_prim_mat = U_mat - V_mat / Q_mat * V_mat';
ones_vec = ones(sum(~miss_ind), 1);

% compute aggregated prediction
aggreg_pred = expert_preds(~miss_ind) * U_prim_mat * ones_vec / sum(U_prim_mat(:));