function [Sigma, q, Lambda] = semisuper_combination_of_experts_train(truth_train, Y_train, X_train, R)

% function [Sigma, q, Lambda] = semisuper_combination_of_experts_train(truth_train, Y_train, X_train, R)
%
% Function trains a regression model for semi-supervised combination of 
% experts. Given ground truth labels, experts' predictions, explanatory
% variables for each training label, and expected number of data-generating
% regimes, it outputs the parameters of the aggregation model.
%
% Inputs: 
%   truth_train - vector of length N, holding ground-truth labels; NaN 
%                 denotes missing label
%   Y_train     - (N x K) matrix, each row represents predictions of K 
%                 experts for one training example; NaN denotes missing
%                 expert prediction
%   X_train     - (N x D) matrix, each row represents explanatory variables
%                 for the corresponding training example; if R is set to
%                 number larger than 1, value of D needs to be larger than 
%                 0, otherwise if R = 1 you can leave this input empty
%   R           - number of regimes that generated the training data
%
% Outputs: 
%   Sigma       - cell array of length R, each element is the
%                 sigma matrix for one regime; see eq. (2) in the paper
%   q           - (R x D) matrix, each row is a prototype vector for one
%                 regime; see eq. (28)
%   Lambda      - cell array of length R, each element is a feature scaling
%                 matrix for one regime; see eq. (28)
%  
% Author: Nemanja Djuric (nemanja@temple.edu), according to paper Djuric et
%         al., 'Semi-Supervised Learning for Integration of Aerosol
%         Predictions from Multiple Satellite Instruments', IJCAI 2013.
% Date  : October 25th, 2013
%
% Update on April 7th, 2014: We corrected the end of M-step, where gradient update
%							 for Lambda matrix should be *subtracted* and not *added* 
% 							 as stated in eq. (31) from the paper. Both this 
%							 published code and paper had typos, be aware. We stress
%							 that in the paper experiments we used the correct version.
% Update on May 15th, 2014: Minor fixes of computation of p_r for labeled
%                           and unlabeled points.

%% modifiable parameters of the algorithm (GD - gradient descent)
[N, K] = size(Y_train);     % N - training size, K - number of experts
learning_rate_q = 0.1;      % learning rate for GD learning of q vectors
learning_rate_Lambda = 0.01;% learning rate for GD learning of Lambda mats
prior_cov_matrix = eye(K);  % prior variance of experts; see Section 2.4;
                            %   here we assume diagonal prior although if
                            %   you want change it to any valid cov. matrix
num_em_iters = 15;          % number of EM iterations; see Section 2.5
num_Sigma_iters = 50;       % number of iterations to find sigma matrices;
                            %   see Sections 2.2 and/or 2.3

% we put unlabeled points first, labeled at the end, as in the paper
ind_labeled = ~isnan(truth_train);
Y_train = [Y_train(~ind_labeled, :); Y_train(ind_labeled, :)];
truth_train = [truth_train(~ind_labeled); truth_train(ind_labeled)];
missing_ind_train = isnan(Y_train);
% get sizes of unlabeled and labaled training set
N_u = sum(~ind_labeled);
N_l = sum(ind_labeled);

% if X_train is empty then just make it a vector of zeros, it will make all
% p_ir's below to have the same value
if (isempty(X_train)) || (nargin < 3)
    R = 1;
    num_em_iters = 1;
    num_features = 1;
    X_train = zeros(N, num_features);
else
    num_features = size(X_train, 2);
    X_train = [X_train(~ind_labeled, :); X_train(ind_labeled, :)];
end;

if (nargin < 4) || (R == 1)
    % by default set to one cluster, then we also don't need any EM iters
    R = 1;
    num_em_iters = 1;
end;

% initialize parameters of the model, modify if needed
Sigma = cell(1, R);
Lambda = cell(1, R);
for r = 1 : R
    Sigma{r} = eye(K);              % initial guesses are identity matrices
    Lambda{r} = eye(num_features);
end;
q = 0.5 * randn(R, num_features); % random initial guess for q vectors

%% TRAINING PHASE STARTS HERE
% note: remember first N_u points are unlabeled, last N_l points labeled
for em_iteration = 1 : num_em_iters
    %% E-step:
    % compute logarithm of the determinant of all sigma matrices to speed
    % up computation later
    log_det_Sigma = zeros(1, R);
    Sigma_inv = cell(1, R);
    for r = 1 : R
        Sigma_inv{r} = inv(Sigma{r});
        log_det_Sigma(r) = log(det(Sigma{r}));
    end;
    
    % compute pi_ir for all points
    pi_ir = zeros(N, R);
    for i = 1 : N
        for r = 1 : R
            pi_ir(i, r) = exp(-(X_train(i, :) - q(r, :)) * Lambda{r} * (X_train(i, :) - q(r, :))');
        end;
        pi_ir(i, :) = pi_ir(i, :) / sum(pi_ir(i, :));
    end;
    
    % compute h_ir on unlabeled data
    h_ir = zeros(N, R);
    for i = 1 : N_u
        miss_ind = missing_ind_train(i, :); 
        if (sum(miss_ind) == K)
            error('Error, all experts are missing for a training example!');
        end;
        
        non_miss_ind = ~miss_ind;
        for r = 1 : R            
            % integrate out the missing data
            Q_mat = Sigma_inv{r}(miss_ind, miss_ind);
            V_mat = Sigma_inv{r}(non_miss_ind, miss_ind);
            ones_vec = ones(sum(non_miss_ind), 1);
            U_mat = Sigma_inv{r}(non_miss_ind, non_miss_ind);
            U_prim_mat = U_mat - (V_mat / Q_mat) * V_mat';
            
            % compute constants
            C1 = 0.5 * sum(U_prim_mat(:));
            C2 = Y_train(i, non_miss_ind) * U_prim_mat * ones_vec;
            C3 = 0.5 * Y_train(i, non_miss_ind) * U_prim_mat * Y_train(i, non_miss_ind)';
            
            p_r = -0.5 * log_det_Sigma(r) - 0.5 * log(det(Q_mat)) - 0.5 * log(C1) + C2^2 / (4 * C1) - C3;
            h_ir(i, r) = pi_ir(i, r) * exp(p_r);
            
            if (~isreal(h_ir(i, r)))
                error('Something is wrong, your expectations are imaginary! :)');
            end;
        end;
        h_ir(i, :) = h_ir(i, :) / sum(h_ir(i, :));
    end;
    % compute h_ir on labeled data
    for i = N_u + 1 : N
        miss_ind = missing_ind_train(i, :);          
        if (sum(miss_ind) == K)
            error('Error, all experts are missing for a training example!');
        end;
        
        non_miss_ind = ~miss_ind;
        for r = 1 : R
            % integrate out the missing data
            Q_mat = Sigma_inv{r}(miss_ind, miss_ind);
            V_mat = Sigma_inv{r}(non_miss_ind, miss_ind);
            U_mat = Sigma_inv{r}(non_miss_ind, non_miss_ind);
            U_prim_mat = U_mat - (V_mat / Q_mat) * V_mat';
            
            p_r = 0.5 * log(det(U_prim_mat)) ...
                - 0.5 * (Y_train(i, non_miss_ind) - truth_train(i)) * U_prim_mat * (Y_train(i, non_miss_ind) - truth_train(i))';
            h_ir(i, r) = pi_ir(i, r) * exp(p_r);
            
            if (~isreal(h_ir(i, r)))
                error('Something is wrong, your expectations are imaginary! :)');
            end;
        end;
        h_ir(i, :) = h_ir(i, :) / sum(h_ir(i, :));
    end;
    
    %% M-step, we first find Sigma matrices
    for iter = 1 : num_Sigma_iters
        Sigma_inv = cell(1, R);
        Sigma_gradient = cell(1, R);
        for r = 1 : R
            Sigma_inv{r} = inv(Sigma{r});
            Sigma_gradient{r} = zeros(K);
        end;
        
        %% compute the gradient on unlabeled points
        for i = 1 : N_u
            miss_ind = missing_ind_train(i, :);   
            non_miss_ind = ~miss_ind;
            
            for r = 1 : R
                % integrate out the missing data
                Q_mat = Sigma_inv{r}(miss_ind, miss_ind);
                V_mat = Sigma_inv{r}(non_miss_ind, miss_ind);
                ones_vec = ones(sum(non_miss_ind), 1);
                U_mat = Sigma_inv{r}(non_miss_ind, non_miss_ind);
                V_mat_div_Q_mat = V_mat / Q_mat;
                U_prim_mat = U_mat - V_mat_div_Q_mat * V_mat';

                % compute constants
                C1 = 0.5 * sum(U_prim_mat(:));
                C2 = Y_train(i, non_miss_ind) * U_prim_mat * ones_vec;

                % now account for all parts of equation (30) in the paper
                temp = zeros(K);
                inv_Q_mat = inv(Q_mat);
                % part 1:
                temp(miss_ind, miss_ind) = inv_Q_mat;
                % part 2 and 4b:
                temp = temp + (1 / (2 * C1) + C2^2 / (4 * C1^2)) * get_missing_data_matrix(ones(sum(non_miss_ind)), V_mat_div_Q_mat, miss_ind);
                % part 3:
                temp = temp + get_missing_data_matrix(Y_train(i, non_miss_ind)' * Y_train(i, non_miss_ind), V_mat_div_Q_mat, miss_ind);
                % part 4a:
                temp = temp - 0.5 * C2 / C1 * get_missing_data_matrix(ones_vec * Y_train(i, non_miss_ind) + Y_train(i, non_miss_ind)' * ones_vec', V_mat_div_Q_mat, miss_ind);
                
                % add everything to gradient
                Sigma_gradient{r} = Sigma_gradient{r} + h_ir(i, r) * temp;
            end;
        end;
        
        %% compute the gradient on labeled points
        for i = N_u + 1 : N
            miss_ind = missing_ind_train(i, :);
            non_miss_ind = ~miss_ind;
            
            for r = 1 : R
                % integrate out the missing data
                Q_mat = Sigma_inv{r}(miss_ind, miss_ind);
                V_mat = Sigma_inv{r}(non_miss_ind, miss_ind);
                V_mat_div_Q_mat = V_mat / Q_mat;

                % now account for all parts of equation (30) in the paper
                temp = zeros(K);
                % part 1:
                temp(miss_ind, miss_ind) = inv(Q_mat);
                % part 2:
                temp = temp + get_missing_data_matrix((Y_train(i, non_miss_ind) - truth_train(i))' * (Y_train(i, non_miss_ind) - truth_train(i)), V_mat_div_Q_mat, miss_ind);
                
                % add everything to gradient
                Sigma_gradient{r} = Sigma_gradient{r} + h_ir(i, r) * temp;
            end;
        end;
        % compute new sigma
        for r = 1 : R
            Sigma{r} = (Sigma_gradient{r} + prior_cov_matrix) / (sum(h_ir(:, r)) + (sum(prior_cov_matrix(:)) ~= 0));
        end;
    end;
    
    %% continuing M-step
    % first find weight vectors gradient, ...
    gradient_q = zeros(R, num_features);
    for i = 1 : N
        for r = 1 : R
            gradient_q(r, :) = gradient_q(r, :) + (h_ir(i, r) - pi_ir(i, r)) * (X_train(i, :) - q(r, :)) * Lambda{r}';
        end;
    end;
    % ..., then find scaling matrix gradient
    gradient_Lambda = cell(1, R);
    for r = 1 : R
        gradient_Lambda{r} = zeros(num_features);
        for i = 1 : N
            gradient_Lambda{r} = gradient_Lambda{r} + (h_ir(i, r) - pi_ir(i, r)) * (X_train(i, :) - q(r, :))' * (X_train(i, :) - q(r, :));
        end;        
        Lambda{r} = Lambda{r} - learning_rate_Lambda * gradient_Lambda{r};
    end;    
    q = q + learning_rate_q * gradient_q;
    
    % print info; if you do not want to see how the parameters are changing
    % during learning comment this part
    em_iteration
    q
    for r = 1 : R
        temp_diag = sqrt(diag(Sigma{r}))'
    end;
end;
% figure, scatter(X_train(:, 1), X_train(:, 2), 50, (pi_ir(:, 1)>0.5), 'filled')

function return_mat = get_missing_data_matrix(input_mat, V_mat_div_Q_mat, miss_ind)
% here we compute the matrix from eq. (18) from the IJCAI paper
return_mat = zeros(length(miss_ind));
return_mat(~miss_ind, ~miss_ind) = input_mat;
return_mat(~miss_ind, miss_ind) = -input_mat' * V_mat_div_Q_mat;
return_mat(miss_ind, ~miss_ind) = return_mat(~miss_ind, miss_ind)';
return_mat(miss_ind, miss_ind) = V_mat_div_Q_mat' * input_mat * V_mat_div_Q_mat;